'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=no-member
# pylint: disable=superfluous-parens
# pylint: disable=too-many-arguments
import re
import uuid
from xml.etree import ElementTree

from synbiochem.utils.dna_utils import DNA


_RDF_NS = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'
_NS = {'ns': 'http://sbols.org/v1#',
       'rdf': _RDF_NS}


def read(filename):
    '''Parses SBOL v1.'''
    tree = ElementTree.parse(filename)
    root = tree.getroot()

    dna_comp = root.find('ns:DnaComponent', _NS)
    dna_seq = dna_comp.find('ns:dnaSequence', _NS)
    dna_seq = dna_seq.find('ns:DnaSequence', _NS)

    params = _read_dna_comp(dna_comp)
    params.update({'seq': dna_seq.find('ns:nucleotides', _NS).text})
    dna = DNA(**params)

    for annot in dna_comp.findall('ns:annotation', _NS):
        _read_annot(dna, annot)

    return dna


def write(dna, filename=None):
    '''Writes a Dna object to SBOL v1.'''
    root = ElementTree.Element('ns2:RDF', {'xmlns': 'http://sbols.org/v1#',
                                           'xmlns:ns2': _RDF_NS})

    dna_comp = _write_dna_comp(root, dna)

    dna_seq = _write(dna_comp, 'dnaSequence')
    dna_seq = _write(dna_seq, 'DnaSequence', _get_about())
    _write(dna_seq, 'nucleotides', text=dna['seq'])

    for feature in dna['features']:
        annot = _write(dna_comp, 'annotation')
        annot = _write(annot, 'SequenceAnnotation', _get_about())
        _write(annot, 'bioStart', text=str(feature['start']))
        _write(annot, 'bioEnd', text=str(feature['end']))
        _write(annot, 'strand', text='+' if feature['forward'] else '-')
        sub_comp = _write(annot, 'subComponent')
        _write_dna_comp(sub_comp, feature)

    sbol = ElementTree.tostring(root)

    with open(filename, 'wb') as outfile:
        outfile.write(sbol)

    return sbol


def _read_dna_comp(dna_comp):
    '''Read DNAComponent node.'''
    disp_id = _find_text(dna_comp, 'ns:displayId')
    name = _find_text(dna_comp, 'ns:name')
    desc = _find_text(dna_comp, 'ns:description')
    typ_node = dna_comp.find('rdf:type', _NS)
    typ = typ_node.attrib['{' + _RDF_NS + '}resource'] \
        if typ_node is not None else None

    if not re.match('^[\\w-]+$', disp_id):
        disp_id = str(uuid.uuid4())

    return {'disp_id': disp_id, 'name': name, 'desc': desc, 'typ': typ}


def _read_annot(dna, annot):
    '''Reads annotation node.'''
    seq_annot = annot.find('ns:SequenceAnnotation', _NS)
    sub_comp = seq_annot.find('ns:subComponent', _NS)
    dna_comp = sub_comp.find('ns:DnaComponent', _NS)

    params = _read_dna_comp(dna_comp)

    start = _find_text(seq_annot, 'ns:bioStart')

    if start:
        params.update({'start': int(start)})

    end = _find_text(seq_annot, 'ns:bioEnd')

    if end:
        params.update({'end': int(end)})

    forward = _find_text(seq_annot, 'ns:strand')

    if forward:
        params.update({'forward': forward == '+'})

    # Tests due to ICE eccentricities...
    try:
        feat = DNA(**params)

        pos = (feat['start'], feat['end'], feat['forward'])

        # Prevents cases where features are duplicated in the SBOL:
        if pos not in [(feature['start'], feature['end'], feature['forward'])
                       for feature in dna['features']]:
            dna['features'].append(feat)

    except ValueError:
        # Prevents cases with no end position and no sequence:
        print('Ignoring invalid feature.')


def _find_text(parent, field):
    '''Finds text from node.'''
    node = parent.find(field, _NS)
    return None if node is None else node.text


def _write_dna_comp(parent, dna):
    '''Write DNAComponent node.'''
    dna_comp = ElementTree.SubElement(parent, 'DnaComponent',
                                      _get_about(dna['disp_id']))

    if dna['typ']:
        _write(dna_comp, 'ns2:type', {'ns2:resource': dna['typ']})

    _write(dna_comp, 'displayId', text=dna['disp_id'])

    if dna['name']:
        _write(dna_comp, 'name', text=dna['name'])

    if dna['desc']:
        _write(dna_comp, 'description', text=dna['desc'])

    return dna_comp


def _write(parent, name, params=None, text=None):
    '''Writes a node.'''
    if params is None:
        params = {}

    node = ElementTree.SubElement(parent, name, params)
    node.text = text
    return node


def _get_about(uid=None):
    '''Gets about attributes.'''
    if uid is None:
        uid = str(uuid.uuid4())

    return {'ns2:about': 'https://www.synbiochem.co.uk#' + uid}



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=no-member
# pylint: disable=too-many-public-methods
import os
import tempfile
import unittest

from synbiochem.utils import sbol_utils


class Test(unittest.TestCase):
    '''Test class for dna_utils.'''

    def test(self):
        '''Tests round trip equality.'''
        directory = os.path.dirname(os.path.realpath(__file__))
        dna1 = sbol_utils.read(os.path.join(directory, 'sbol.xml'))
        dna2 = round_trip(dna1)
        self.assertEqual(dna1, dna2)


def round_trip(dna):
    '''Writes / reads DNA object, via SBOL export / import.'''
    tmp = tempfile.NamedTemporaryFile()
    sbol_utils.write(dna, tmp.name)
    return sbol_utils.read(tmp.name)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=broad-except
# pylint: disable=too-many-arguments
import math
import random
import traceback

from synbiochem.utils.job import JobThread


class SimulatedAnnealer(JobThread):
    '''Class to perform simulated annealing method.'''

    def __init__(self, solution,
                 acceptance=0.01, max_iter=10000,
                 r_temp=0.025, cooling_rate=2.5e-4,
                 heartbeat=1, verbose=False):
        self.__solution = solution
        self.__acceptance = acceptance
        self.__max_iter = max_iter
        self.__r_temp = r_temp
        self.__cooling_rate = cooling_rate
        self.__verbose = verbose
        self.__heartbeat = heartbeat
        JobThread.__init__(self)

    def run(self):
        '''Optimises a solution with simulated annealing.'''
        try:
            if self.__init():
                # Initialization:
                iteration = 0
                accepts = 0
                rejects = 0
                r_temp = self.__r_temp

                energy = self.__solution.get_energy()

                while not self._cancelled \
                        and energy > self.__acceptance \
                        and iteration < self.__max_iter:
                    iteration += 1
                    energy_new = self.__solution.mutate()

                    if energy_new < energy:
                        # Accept move immediately:
                        self.__accept(iteration)
                        self.__log(iteration, r_temp, energy, energy_new, 'T')
                        energy = energy_new
                    elif energy_new == energy:
                        # Reject move:
                        self.__solution.reject()
                        rejects += 1
                    elif _get_exp((energy - energy_new) / r_temp) > \
                            random.random():
                        # Accept move based on conditional probability:
                        self.__accept(iteration)
                        self.__log(iteration, r_temp, energy, energy_new, '*')
                        energy = energy_new
                        accepts += 1
                    else:
                        # Reject move:
                        self.__solution.reject()
                        rejects += 1

                    # Heartbeat:
                    self.__check_heartbeat(iteration)

                    # Simulated annealing control:
                    accepts, rejects, r_temp = \
                        _control_r_temp(accepts, rejects, r_temp)

                    r_temp *= 1 - self.__cooling_rate

                self.__log(iteration, r_temp, energy, energy_new, 'E')
                self.__exit(iteration)
        except Exception:
            self.__fire_event('error', 100, 0, message=traceback.format_exc())

    def __init(self):
        '''Initialise.'''
        self.__fire_event('running', 0, 0, message='Job initialising...')

        try:
            self.__solution.init()
        except ValueError:
            self.__fire_event('error', 100, 0, message=traceback.format_exc())
            return False

        self.__fire_event('running', 0, 0, message='Job initialised')
        return True

    def __log(self, iteration, r_temp, energy, energy_new, symbol):
        '''Log data.'''
        if self.__verbose:
            print('\t'.join([symbol,
                             str(iteration),
                             str(energy),
                             str(energy_new),
                             '%.2f' % r_temp,
                             '%.2f' % _get_exp((energy - energy_new) / r_temp),
                             str(self.__solution)]))

    def __check_heartbeat(self, iteration):
        '''Heartbeat.'''
        if float(iteration) % self.__heartbeat == 0:
            self.__fire_event('running',
                              float(iteration) / self.__max_iter * 100,
                              iteration,
                              'Running...')

    def __accept(self, iteration):
        '''Accept the current solution.'''
        self.__solution.accept()

        self.__fire_event('running', float(iteration) / self.__max_iter * 100,
                          iteration, 'Running...')

    def __fire_event(self, status, progress, iteration, message=''):
        '''Fires an event.'''
        event = {'update': {'status': status,
                            'message': message,
                            'progress': progress,
                            'iteration': iteration,
                            'max_iter': self.__max_iter,
                            'values': self.__solution.get_values()},
                 'query': self.__solution.get_query()
                 }

        if status == 'finished':
            event['result'] = self.__solution.get_result()

        self._fire_event(event)

    def __exit(self, iteration):
        '''Exit.'''
        if iteration == self.__max_iter:
            message = 'Unable to optimise in ' + str(self.__max_iter) + \
                ' iterations'
            self.__fire_event('error', 100, iteration, message=message)
        elif self._cancelled:
            self.__fire_event('cancelled', 100, iteration,
                              message='Job cancelled')
        else:
            self.__fire_event('finished', 100, iteration,
                              message='Job completed')


def _control_r_temp(accepts, rejects, r_temp):
    '''Control temperature.'''
    if accepts + rejects > 50:
        if float(accepts) / float(accepts + rejects) > 0.2:
            # Too many accepts, reduce r_temp:
            r_temp /= 2.0
            accepts = 0
            rejects = 0
        elif float(accepts) / float(accepts + rejects) < 0.025:
            # Too many rejects, increase r_temp:
            r_temp *= 2.0
            accepts = 0
            rejects = 0

    return accepts, rejects, r_temp


def _get_exp(value):
    '''Get exp safely.'''
    try:
        ans = math.exp(value)
    except OverflowError:
        ans = float('inf')

    return ans



'''
SYNBIOCHEM-DB (c) University of Manchester 2015

SYNBIOCHEM-DB is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
from synbiochem import biochem4j


def get_synonyms_by_id(tax_id):
    '''Gets a taxonomy's children.'''
    qry = 'MATCH (o:Organism {taxonomy: {tax_id}}) RETURN o'

    parameters = {'tax_id': tax_id}

    results = _parse(biochem4j.run_query(qry, parameters))
    return results[0]['names'] if results else ''


def get_children_by_id(tax_id):
    '''Gets a taxonomy's children.'''
    qry = 'MATCH (p:Organism {taxonomy: {tax_id}})<--(c:Organism) ' + \
        'RETURN c'

    parameters = {'tax_id': tax_id}

    return _parse(biochem4j.run_query(qry, parameters))


def get_parent_by_id(tax_id):
    '''Gets a taxonomy parent.'''
    qry = 'MATCH (c:Organism {taxonomy: {tax_id}})-->(p:Organism) ' + \
        'RETURN p'

    parameters = {'tax_id': tax_id}

    results = _parse(biochem4j.run_query(qry, parameters))
    return results[0]


def get_parent_by_name(tax_name):
    '''Gets a taxonomy parent.'''

    qry = 'MATCH (c:Organism)-->(p:Organism) WHERE \'' + \
        tax_name + \
        '\' IN c.names RETURN p'

    parameters = {}

    results = _parse(biochem4j.run_query(qry, parameters))
    return results[0]


def _parse(data):
    '''Parses data.'''
    if data['errors']:
        raise ValueError(str(data['errors']))

    return [datum['row'][0] for datum in data['results'][0]['data']]



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
from threading import Thread
import uuid


class JobThread(Thread):
    '''Wraps a job into a thread, and fires events.'''

    def __init__(self):
        Thread.__init__(self)
        self._cancelled = False
        self.__job_id = str(uuid.uuid4())
        self.__listeners = set()

    def get_job_id(self):
        '''Gets thread job id.'''
        return self.__job_id

    def cancel(self):
        '''Cancels the current job.'''
        self._cancelled = True

    def add_listener(self, listener):
        '''Adds an event listener.'''
        self.__listeners.add(listener)

    def remove_listener(self, listener):
        '''Removes an event listener.'''
        self.__listeners.remove(listener)

    def _fire_event(self, event):
        '''Event listener, passes events on to registered listeners.'''
        event.update({'job_id': self.__job_id})

        for listener in self.__listeners:
            listener.event_fired(event)



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=relative-import
import gzip
import os
import tempfile
import zipfile

from six.moves.urllib import request


def get_file(source_url, target_filename):
    '''Downloads file from url and saves to target.'''

    if not os.path.isfile(target_filename):
        if not os.path.exists(os.path.dirname(target_filename)):
            os.makedirs(os.path.dirname(target_filename))

        request.urlretrieve(source_url, target_filename)

    destination = os.path.dirname(target_filename)

    if target_filename.endswith('.zip'):
        zfile = zipfile.ZipFile(target_filename, 'r')
        target_filename = os.path.join(destination, zfile.namelist()[0])
        zfile.extractall(destination)
    elif target_filename.endswith('.gz'):
        unzipped_filepath = target_filename[:-len('.gz')]

        if os.path.exists(unzipped_filepath):
            target_filename = unzipped_filepath
        else:
            input_file = gzip.open(target_filename, 'rb')
            target_filename = os.path.join(destination,
                                           input_file.name[:-len('.gz')])
            output_file = open(target_filename, 'wb')

            for line in input_file:
                output_file.write(line)

            input_file.close()
            output_file.close()

    return target_filename


def get_filename(filename):
    '''Returns a filename, generating a temp file if necessary.'''
    if filename is None:
        fle = tempfile.NamedTemporaryFile(delete=False)
        return fle.name

    return filename



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=no-self-use
# pylint: disable=protected-access
# pylint: disable=too-few-public-methods
# pylint: disable=too-many-arguments
import random

import numpy


class Chromosome():
    '''Class to represent a chromosome.'''

    def __init__(self, chromosome, mutation_rate=0.1):
        self._chromosome = chromosome
        self.__mutation_rate = mutation_rate

    def get_chromosome(self):
        '''Gets the chromosome.'''
        return self._chromosome

    def mutate(self):
        '''Mutate.'''
        return None

    def breed(self, partner):
        '''Breed.'''
        pass

    def fitness(self):
        '''Get fitness.'''
        pass

    def __repr__(self):
        return str(self._chromosome)


class GeneticAlgorithm():
    '''Base class to run a genetic algorithm.'''

    def __init__(self, population, retain=0.05, random_select=0.2,
                 mutate=0.25, verbose=False):
        self.__population = population
        self.__retain = retain
        self.__random_select = random_select
        self.__mutate = mutate
        self._verbose = verbose

    def run(self, max_iter=1024):
        '''Run.'''
        for _ in range(max_iter):
            results = self.__evolve()

            if results is not None:
                return results

        raise ValueError('Unable to optimise in %d iterations.' % max_iter)

    def get_results(self):
        '''Get results.'''
        return sorted([(chrm.fitness(), chrm) for chrm in self.__population])

    def __evolve(self):
        results = self.get_results()

        if results[0][0] == 0:
            return results

        if self._verbose:
            print(results[0])

        results = [x[1] for x in results]
        retain_length = int(len(self.__population) * self.__retain)

        # Retain best and randomly add other individuals to promote genetic
        # diversity:
        new_population = results[:retain_length]

        # Mutate some individuals:
        for individual in new_population:
            if self.__mutate > random.random():
                mutated = individual.mutate()

                if mutated:
                    new_population.append(mutated)

        new_population.extend([ind for ind in results[retain_length:]
                               if self.__random_select > random.random()])

        while len(new_population) < len(self.__population):
            male = random.choice(self.__population)
            female = random.choice(self.__population)

            if male != female:
                child = male.breed(female)

                if child:
                    new_population.append(child)

                new_population = list(
                    numpy.unique(numpy.array(new_population)))

        self.__population = new_population

        # print numpy.mean([self._fitness(ind) for ind in self.__population])

        return None



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=invalid-name
# pylint: disable=wrong-import-order
import unittest

from synbiochem.utils import neo4j_utils

import pandas as pd


class Test(unittest.TestCase):
    '''Test class for neo4j_utils.'''

    def test_type_files(self):
        '''Tests type_files method.'''
        data = [[[1.0, 3.7, -34.8e-17],
                 ['some', 'random', 'string'],
                 [4, -3],
                 [True, False, True],
                 32.124,
                 -34,
                 'random string',
                 False,
                 13,
                 56.3,
                 [23.5, 'Tree', False]],
                [[1.6, 3.77, -374.8e-17],
                 ['another', 'random', 'string'],
                 [47, -31],
                 [True, True, False, True],
                 132.124,
                 -314,
                 'another random string',
                 True,
                 34,
                 78.3,
                 [-123.5, 6, 'Trees', True]]]

        columns = ['float_array',
                   'string_array',
                   'int_array',
                   'boolean_array',
                   'float',
                   'int',
                   'string',
                   'boolean',
                   ':START_ID',
                   ':END_ID(Label)',
                   ':LABEL']

        df = pd.DataFrame(data, columns=columns)
        new_df = neo4j_utils.type_df(df, array_delimiter='|')

        expected = ['float_array:float[]',
                    'string_array:string[]',
                    'int_array:int[]',
                    'boolean_array:boolean[]',
                    'float:float',
                    'int:int',
                    'string:string',
                    'boolean:boolean',
                    ':START_ID',
                    ':END_ID(Label)',
                    ':LABEL']

        self.assertEqual(sorted((list(new_df.columns))), sorted(expected))


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''


class Vertex():
    '''Class to represent a vertex.'''

    def __init__(self, name, attributes):
        self.__attributes = attributes
        self.__attributes['name'] = name
        self.__in = []
        self.__out = []

    def predecessors(self):
        '''Get predecessors.'''
        return self.__in

    def attributes(self):
        '''Get attributes.'''
        return self.__attributes

    def add_out(self, vertex_to, attributes):
        '''Add edge out.'''
        self.__out.append((vertex_to, attributes))

    def add_in(self, vertex_from, attributes):
        '''Add edge in.'''
        self.__in.append((vertex_from, attributes))

    def indegree(self):
        '''Get indegree.'''
        return len(self.__in)

    def is_root(self):
        '''Is the Vertex a root?'''
        return not self.__out

    def __repr__(self):
        return self.__attributes['name']


class Graph():
    '''Class to represent a graph.'''

    def __init__(self):
        self.__vertices = {}

    def find_vertex(self, name):
        '''Find vertex.'''
        if name in self.__vertices:
            return self.__vertices[name]

        raise ValueError(name)

    def add_vertex(self, name, attributes):
        '''Add vertex.'''
        self.__vertices[name] = Vertex(name, attributes)

    def add_edge(self, vertex_from, vertex_to, attributes):
        '''Add edge.'''
        vertex_from.add_out(vertex_to, attributes)
        vertex_to.add_in(vertex_from, attributes)

    def get_roots(self):
        '''Get roots.'''
        return [vtx for vtx in self.__vertices.values() if vtx.is_root()]


def add_vertex(graph, name, attributes=None):
    '''Add vertex.'''
    try:
        return graph.find_vertex(name)
    except ValueError:
        if not attributes:
            attributes = {}

        graph.add_vertex(name, attributes)
        return graph.find_vertex(name)


def add_edge(graph, vertex_from, vertex_to, attributes=None):
    '''Add edge.'''
    if not attributes:
        attributes = {}

    graph.add_edge(vertex_from, vertex_to, attributes)


def get_roots(graph):
    '''Get roots.'''
    return graph.get_roots()



'''
synbiochem (c) University of Manchester 2015

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
from setuptools import find_packages, setup


setup(name='synbiochem-py',
      version='0.6.23',
      description='synbiochem-py: Core python modules for SYNBIOCHEM',
      long_description='synbiochem-py: Core python modules for SYNBIOCHEM',
      url='https://github.com/synbiochem/synbiochem-py',
      author='Neil Swainston',
      author_email='neil.swainston@liverpool.ac.uk',
      license='MIT',
      classifiers=[
                   'Development Status :: 4 - Beta',
                   'Intended Audience :: Developers',
                   'Topic :: Software Development :: Build Tools',
                   'License :: OSI Approved :: MIT License',
                   'Programming Language :: Python :: 3.7'
      ],
      keywords='synbio synthetic biology',
      packages=find_packages(),
      include_package_data=True,
      test_suite='synbiochem.utils.test',
      install_requires=['biopython',
                        'future',
                        'numpy',
                        'pandas',
                        'requests[security]'])



'''
SYNBIOCHEM-DB (c) University of Manchester 2015

SYNBIOCHEM-DB is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
import json

from synbiochem.utils import net_utils


def run_query(query, parameters):
    '''Gets a reaction.'''
    url = 'http://www.biochem4j.org/db/data/transaction/commit'

    data = {'statements': [
        {
            'statement': query,
            'parameters': parameters
        }
    ]}

    headers = {'Accept': 'application/json; charset=UTF-8',
               'Content-Type': 'application/json'}

    return json.loads(net_utils.post(url, json.dumps(data), headers))



'''
synbiochem (c) University of Manchester 2016

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''


def get_well(idx, rows=8, columns=12, col_ord=True):
    '''Map idx to well'''
    if idx < 0 or idx >= rows * columns:
        raise ValueError('Index %idx out of range' % idx)

    return _get_well_col(idx, rows) if col_ord \
        else _get_well_row(idx, columns)


def _get_well_col(idx, rows):
    '''Map idx to well, column ordered.'''
    return chr(ord('A') + (idx % rows)) + str(idx / rows + 1)


def _get_well_row(idx, columns):
    '''Map idx to well, row ordered.'''
    return chr(ord('A') + (idx / columns)) + str(idx % columns + 1)



'''
synbiochem (c) University of Manchester 2017

synbiochem is licensed under the MIT License.

To view a copy of this license, visit <http://opensource.org/licenses/MIT/>.

@author:  neilswainston
'''
# pylint: disable=invalid-name
# pylint: disable=too-many-arguments
import functools
import os
import shutil
import subprocess

import numpy as np
import pandas as pd


def create_db(neo4j_root_loc, nodes_files, rels_files,
              delimiter, array_delimiter,
              db_name='graph.db',
              ignore_duplicates=False):
    '''Creates the database from csv files.

    Import format is:

    neo4j-admin import --nodes NODES_FILE --relationships RELS_FILE

    '''
    # Apply types to import files:
    for filename in nodes_files + rels_files:
        df = pd.read_csv(filename, sep=delimiter,
                         low_memory=False, dtype=object)
        df = type_df(df, array_delimiter)
        df.to_csv(filename, delimiter, encoding='utf-8', index=False)

    # Stop database:
    bin_loc = os.path.join(neo4j_root_loc, 'bin')
    subprocess.call([os.path.join(bin_loc, 'neo4j'), 'stop'])

    # Remove database directory if it already exists:
    db_loc = os.path.join(neo4j_root_loc, 'data', 'databases', db_name)

    if os.path.exists(db_loc):
        shutil.rmtree(db_loc)

    # Import database:
    admin_loc = os.path.join(bin_loc, 'neo4j-admin')

    params = [admin_loc,
              'import',
              '--database', db_name]

    for node_file in nodes_files:
        params.extend(['--nodes', node_file])

    for rels_file in rels_files:
        params.extend(['--relationships', rels_file])

    if ignore_duplicates:
        params.append('--ignore-duplicate-nodes')

    # not sure why, but this bit is required by ENI model
    # seems like one more bug regarding reading csv
    params.extend(['--multiline-fields', 'True'])

    params.extend(['--delimiter', delimiter])
    params.extend(['--array-delimiter', array_delimiter])

    subprocess.call(params)


def type_df(df, array_delimiter):
    '''Format dataframe and update column names in file based on dtype.'''
    # Id / special case columns regexps:
    _format_list_delim = functools.partial(_format_list,
                                           array_delimiter=array_delimiter)
    new_df = pd.DataFrame()

    for col in df:
        if ':' in col:
            new_df[col] = df[col]
        else:
            # Check if list and reformat if so:
            res = df[col].apply(_format_list_delim)
            dtype = _get_type(res[0].unique())

            if dtype:
                new_df[col + ':' + dtype + '[]'] = res[1]
            else:
                # Convert appropriate cols to int / float / boolean:
                try:
                    dtype = _get_type(str(df[col].dtypes))
                    new_df[col.replace(':', '_') + ':' + dtype] = res[1]
                except (TypeError, ValueError):
                    # Ignore errors if data is not numerical.
                    pass

    return new_df


def _format_list(cell, array_delimiter):
    '''Format cell if of type list.'''
    if not isinstance(cell, list):
        return pd.Series([None, cell])

    return pd.Series([pd.Series(cell).dtype,
                      array_delimiter.join(map(str, cell))])


def _get_type(dtype):
    '''Convert dtype to neo4j import type.'''
    if isinstance(dtype, np.ndarray):
        if len(dtype) == 1:
            return _get_type(str(dtype[0]))

        return None

    if dtype == 'bool':
        return 'boolean'
    if dtype == 'object':
        return 'string'
    if dtype in ['nan', 'None']:
        return None

    return ''.join([c for c in dtype if not c.isdigit()])
